# /**
#
# \name{detectReplicas}
#
# \title{detectReplicas}
#
# \alias{detectReplicas}
#
# \keyword{}
#
# \description{}
#
# \arguments{
#  \item{arrayDescription}{data frame which contains a column \code{spotIdentifier}}
#  \item{spotIdentifier}{character string}
#  \item{identifiersToBeSkipped}{vector of character strings to be ignored}
# }
#
# \value{a list containing \code{nrOfReplicas}, i.e. an integer characterizing the
#  same number of spot replicas given for each identifier or otherwise \code{NA}, and
#  containing \code{spotReplicas}, i.e. a list of the length of unique identifiers
#  where each element contains a vector of indeces corresponding to the identifier 
#  used as name for the list element 
#  otherwise \code{NA}}
#
# \examples{
#
#aD <- data.frame(ID=c("Blank", "Blank", "x", "x", "Control1", "y", "x", "z", "y", "Control2"))
#re <- detectReplicas(aD, identifiersToBeSkipped = c("Blank", "Control1", "Control2"))
#stopifnot(is.na(re$nrOfReplicas))
#
#aD <- data.frame(ID=c("Blank", "x", "x", "Control1", "y", "z", "z", "y", "Control2"))
#re <- detectReplicas(aD, identifiersToBeSkipped = c("Blank", "Control1", "Control2"))
#stopifnot(re$nrOfReplicas == 2 )
#
#aD <- data.frame(ID=c("z", "x", "x", "x", "y", "z", "z", "y", "y"))
#re <- detectReplicas(aD, identifiersToBeSkipped = c("Blank", "Control1", "Control2"))
#stopifnot(re$nrOfReplicas == 3 )
#
#aD <- data.frame(ID=c("z", "x", "Control1", "y"))
#re <- detectReplicas(aD, identifiersToBeSkipped = c("Control1"))
#stopifnot(re$nrOfReplicas == 1 )
#
#
#aD <- data.frame(ID=c("Blank", "Control1", "Blank", "Control2"))
#re <- detectReplicas(aD, identifiersToBeSkipped = c("Blank", "Control1", "Control2"))
#stopifnot(is.na(re$nrOfReplicas))
#
# }
#
# \author{Andreas Buness <a.buness@dkfz.de>}
#
# */
detectReplicas <- function( arrayDescription,
                               spotIdentifier = "ID",
                               identifiersToBeSkipped = "Blank"
                               ){
  
  nrOfReplicas <- NA
  spotReplicas <- NA
  
  if(is.null(arrayDescription[,spotIdentifier])){
    tmpString <- paste("spotIdentifier",spotIdentifier," is null in data frame arrayDescription")
    cat(tmpString)
    warning(tmpString)
    return(list(nrOfReplicas=nrOfReplicas,spotReplicas=spotReplicas))
  }

  nonValidItems <- sapply(identifiersToBeSkipped, function(x) return(which( arrayDescription[,spotIdentifier] == x)), simplify=TRUE )
  nonValidItems <- na.exclude(as.numeric(unlist(nonValidItems)))


  if( length(nonValidItems) == 0 ){
    spotIDs <- arrayDescription[,spotIdentifier]
  }else{
    spotIDs <- arrayDescription[,spotIdentifier][-nonValidItems]
  }

  ## duplicate/replicate spots ?
  ## condition is not sufficient; e.g. 1+2+3=6, i.e. unique 3 and length 6 thus 6/3=2
  # uniqueItemsNr <- length(unique(spotIDs))
  #if( length(spotIDs) %% uniqueItemsNr == 0 ){
  #  nrOfReplicas <- length(spotIDs) / uniqueItemsNr
  #}

  if(length(spotIDs) > 0 ){
    spotReplicas <- lapply(unique(spotIDs), function(x) return(which( arrayDescription[,spotIdentifier] == x ))) 
    names(spotReplicas) <- unique(spotIDs)

    distributionOfReplicas <- sapply(spotReplicas, length)
    maxReplicas <-  max(distributionOfReplicas)
    minReplicas <-  min(distributionOfReplicas)
    if( maxReplicas == minReplicas ){
      nrOfReplicas <- maxReplicas
      cat("\n every item is spotted ");cat(nrOfReplicas);cat(" times \n") 
    }else{
      cat(" no single/double/triplcate ... spotting has been detected \n")
      cat(paste(" number of spot replicas varies between ", minReplicas, " and ", maxReplicas, "\n",sep=""))
      cat("distribution - first line number of replicas \n")
      cat("               second line number of corresponding ids \n")
      distTable <- table(factor(distributionOfReplicas))
      cat(format(dimnames(distTable)));cat("\n")
      cat(distTable);cat("\n")
    }
  }
    
  return(list(nrOfReplicas=nrOfReplicas,spotReplicas=spotReplicas))
  
}

