# need of more testing !!


# Recall:  there is a relationship between the PIN and the PLATE
#          and the information given in the gpr files
#          probably PIN via Sector
#          probably PLATE via cloneIdentifier

# /**
#
# \name{normalise}
#
# \title{normalisation}
#
# \alias{normalise}
#
# \keyword{}
#
# \description{}
#
# \arguments{
#  \item{arrayDataObject}{ arrayDataObject contains a raw data object, i.e. a
#                 three dimensional array (spots x channels x hybridisations),
#                 where channels must be labbeled as \code{channelNames}}
#  \item{channelNames}{character vector containing 4 items, by default: }
#  \item{subtractBackground}{logical; default: \code{FALSE}}
#  \item{method}{character string; default: \code{"vsn"};
#        possible: \code{"vsn"}, \code{"loess"},
#                  \code{"loessScale"}, \code{"loessQuantile"},
#                  \code{"quantile"}}
#  \item{channelsSeparately}{apply single channel normalisation to each colour channel separately; not used by ratio based normalistion like lo(w)ess; default: \code{FALSE}}
#  \item{hybridisationGroups}{list of vectors of indexes; normalise each group of hybridisation separately;default: missing}
#
# }
#
# \value{object of class normaliseResult}
#
# \author{Andreas Buness <a.buness@dkfz.de>}
#
# */

normalise <- function(arrayDataObject,
                      channelNames = c("Cy3", "Cy3Background",
                                       "Cy5", "Cy5Background"),
                      subtractBackground = FALSE,
                      method = "vsn",
                      channelsSeparately = FALSE,
                      hybridisationGroups){


  if( channelsSeparately == TRUE & ( method == "loess" |
                                     method == "loessScale"  |
                                     method == "loessQuantile" ) ){
    stop(" invalid combination: method=loess/loessScale/loessQuantile, channelsSeparately=TRUE in normalise")
  }
  
  if( missing(arrayDataObject) ){
    stop("object arrayDataObject is missing in normalise")
  }
  
  rawData <- getIntensities(arrayDataObject)
  
  if( !all(dimnames(rawData)[[2]] %in% channelNames) ){
    cat(paste("channelNames:",paste(channelNames,collapse=","),"\n",sep=""))
    stop("channelNames not found in rawData object in normalise")
  }
  if( ! method %in% c("vsn","loess","loessScale","loessQuantile","quantile") ){
    cat(paste("method:",method,"\n",sep=""))
    stop("invalid method specified in normalise")
  }

  nrspot <- dim(rawData)[1]
  nrhyb  <- dim(rawData)[3]
  pqua   <- array(NA, dim=c(nrspot, 2, nrhyb))
  dimnames(pqua) <- list(NULL, c("Cy3Channel", "Cy5Channel"), NULL)

  # background subtraction ?
  if(subtractBackground){
    pqua[,"Cy3Channel",] <- rawData[,"Cy3",]-rawData[,"Cy3Background",]
    pqua[,"Cy5Channel",] <- rawData[,"Cy5",]-rawData[,"Cy5Background",]
  }else{
    pqua[,"Cy3Channel",] <- rawData[,"Cy3",]
    pqua[,"Cy5Channel",] <- rawData[,"Cy5",]
  }


  # very specific option:
  # definition of groups, i.e. subsets of hybridisation/slides
  # to be normalised and analysed separately as group
  if( missing(hybridisationGroups) ){

    hybridisationGroups <- vector(mode="list", length=1)
    hybridisationGroups[[1]] <- 1:dim(rawData)[3]     

  }else{

    if( ! identical(sort(unlist(hybridisationGroups)),(1:dim(rawData)[3])) ){
      stop("not all hybridisations are found in hybridisationGroups in normalise")
    }
       
  }

  # result objects
  normGroupResults <- vector(mode="list", length=length(hybridisationGroups))
  nqua   <- array(NA, dim=c(nrspot, 2, nrhyb))
  dimnames(nqua) <- list(NULL, c("Cy3", "Cy5"), NULL)
  
  for(i in 1:length(hybridisationGroups)){
    group <- hybridisationGroups[[i]]

    x3 <- pqua[,"Cy3Channel",group]
    y5 <- pqua[,"Cy5Channel",group]

    if( method == "vsn" ){

      cat(" vsn normalisation \n")
      
      if( channelsSeparately == FALSE ){
        
        norm <- vsn(cbind(x3,y5), verbose=TRUE)
        normGroupResults[[i]] <- norm
        firstHalf <- dim(exprs(norm))[2]/2
        nx3 <- exprs(norm)[,1:firstHalf,drop=FALSE] 
        ny5 <- exprs(norm)[,(firstHalf+1):(2*firstHalf),drop=FALSE]
        
      }else{

        normX3 <- vsn(x3, verbose=TRUE)
        normY5 <- vsn(y5, verbose=TRUE)
        normGroupResults[[i]] <- list(cy3Channel=normX3,cy5Channel=normY5)
        nx3 <- exprs(normX3)
        ny5 <- exprs(normY5)
        
      }
      
    }else if( method == "quantile" ){

      # no automatic transformation of the data
      cat(" quantile normalisation \n")
      library(limma)
      if( channelsSeparately == FALSE ){

        norm <- normalizeQuantiles(cbind(x3,y5))
        normGroupResults[[i]] <- list(NULL)
        firstHalf <- dim(norm)[2]/2
        nx3 <- norm[,1:firstHalf,drop=FALSE] 
        ny5 <- norm[,(firstHalf+1):(2*firstHalf),drop=FALSE]

        nx3 <- log(nx3)
        ny5 <- log(ny5)
        
      }else{
        
        normX3 <- normalizeQuantiles(x3)
        normY5 <- normalizeQuantiles(y5)
        normGroupResults[[i]] <- list(cy3Channel=list(NULL),cy5Channel=list(NULL))
        nx3 <- normX3
        ny5 <- normY5

        nx3 <- log(nx3)
        ny5 <- log(ny5)
        
      }
        
    }else if( (method == "loess" | method == "loessScale" | method == "loessQuantile" ) ){

      cat(" loess normalisation \n")

      # useful packages: limma, marray* (Yang), SMA (Dudoit,Yang)
      
      # on lima get loess via
      library(modreg)
      library(limma)


      # M and A versus R and G, cf also with reverse conversion below
      
      # M: numeric matrix containing the M-values or log-2
      # expression ratios. Rows correspond to spots and columns to arrays.
      # A: numeric matrix containing the A-values or average
      #    log-2 expression values 
      # taken from the function MA.RG:
      #    R <- log(R, 2)
      #    G <- log(G, 2)
      #    object$M <- as.matrix(R - G)
      #    object$A <- as.matrix((R + G)/2)
      # and therefore
      # log(G,2) <- A - 0.5 M
      # log(R,2) <- A + 0.5 M

      arrayLayout <- NULL
      if( all( c("Block","Row","Column" ) %in% colnames(getSpotAttr(arrayDataObject)) ) ){
        cat(" calling getLayout: assuming GenePix like layout !!! \n")
        arrayLayout <- getLayout(getSpotAttr(arrayDataObject))
      }
        
      #getLayout <- function (gal) 
      #{
      # n <- nrow(gal)
      # lapply(list(ngrid.r = gal$Block[n]/4,ngrid.c = 4,nspot.r = gal$Row[n], 
      # nspot.c = gal$Column[n]), as.integer)
      # }
      
      G <- x3
      R <- y5
      Gb <- array(data=0,dim=dim(G))
      Rb <- array(data=0,dim=dim(R))
      if(!is.null(getWeights(arrayDataObject)) ){
        weights <- getWeights(arrayDataObject)
        rgListObject <- new("RGList",list(R=R,G=G,Rb=Rb,Gb=Gb,weights=weights))
      }else{
        rgListObject <- new("RGList",list(R=R,G=G,Rb=Rb,Gb=Gb))
      }

      # method options:
      # "none", "median", "loess", "printtiploess",
      # "composite" and "robustspline".
      if( ! is.null( arrayLayout ) ){
        cat(" printiploess normalisation \n")
        maListObject <- normalizeWithinArrays(rgListObject ,arrayLayout, method="printtiploess" )
      }else{
        cat(" standard loess normalisation, i.e no printiploess \n")
        maListObject <- normalizeWithinArrays(rgListObject , method="loess" )
      }

      
#      normalizeBetweenArrays may be called, methods: none, scale, quantile

#     `object=MAList, method=="scale"' M-values are scaled to have the same
#     median-absolute values across arrays, A-values are scaled to have
#     the same medians across arrays

#     `object=MAList, method=="quantile"' M- and A-values are normalized so
#     that the single channel intensities R and G have the same
#     empirical distribution across channels and across arrays
      
#      `object=matrix, method=="scale"' If `object' is a plain matrix, then
#     its columns are normalized to have equal median-absolute-values.

#     `object=matrix, method=="quantile"' If `object' is a plain matrix, then
#     its columns are normalized to have equal empirical distributions. 

      
      if( method == "loessScale" ){
        cat(" normalizeBetweenArrays scale is called normalisation group specific on a MAList object\n")
        maListObject <- normalizeBetweenArrays(maListObject,method="scale")
      }
      if( method == "loessQuantile" ){
        cat(" normalizeBetweenArrays quantile is called normalisation group specific on a MAList object\n")
        maListObject <- normalizeBetweenArrays(maListObject,method="quantile")
      }

      normGroupResults[[i]] <- maListObject
      
      M <- maListObject[["M"]]
      A <- maListObject[["A"]]

      nx3 <- A - 0.5 * M      # G
      ny5 <- A + 0.5 * M      # R


      
    }else{
      stop(" unexpected case; unknown method in normalise ")
    }

    
    nqua[,"Cy3",group] <- nx3
    nqua[,"Cy5",group] <- ny5
    
  }

  result <- new("normaliseResult", normData=nqua, normObject=normGroupResults)
  
  return(result)
  
}
