# /**
#
# \name{removeSpots}
#
# \title{removeSpots}
#
# \alias{removeSpots}
#
# \keyword{}
#
# \description{all elements matching the strings in \code{spotsToBeRemoved}
#              are taken out of the object \code{arrayDataObject} and if
#              supplied out of the object \code{exprSetArrayObject} as well}
#
# \value{list containing
#        an object of type \code{arrayData} labelled arrayDataObject
#        and an integer specifying the number of removed items
#        labelled nrOfRemovedItems
#        and possily an object of type \code{exprSetArray}
#        labelled \code{exprSetArrayObject} }
#
#
# \arguments{
#  \item{arrayDataObject}{object of type \code{arrayData}; required; default missing}
#  \item{exprSetArrayObject}{object of type \code{exprSetArray}; optional; default NULL}
#  \item{spotIdentifier}{character strings name of the appropriate column; default "Name"}
#  \item{spotsToBeRemoved}{vector of character strings; default \code{NULL}}
# }
# 
# \details{}
#
# \seealso{\code{}}
#
# \examples{
#  intensities <- array(data=runif(600),dim=c(100,2,3))
#  dimnames(intensities) <- list(NULL, c("Cy3","Cy5"), NULL)
#  arrayDataObject <- new("arrayData", intensities=intensities, spotattr=data.frame(Name=I(rep(c("x","y","Blank","Blank","z"),20))), hybattr=NULL)
#  res <- removeSpots(arrayDataObject, spotsToBeRemoved=c("x","z"))
#  stopifnot( dim(getIntensities(res[["arrayDataObject"]]))[1] == 3*20 )
#  indCy3=1:3
#  indCy5=4:6
#  channels <- matrix( c(indCy3,indCy5), nrow=length(indCy3), byrow=FALSE )
#  colnames(channels) <- c("Cy3","Cy5")
#  exprSetArrayObject <- new("exprSetArray", exprs=matrix(data=runif(600),nrow=100,ncol=6), phenoData=new("phenoData", pData=data.frame(matrix(0,nrow=6,ncol=1)), varLabels=list(rep("varLabel1",1))), channels=channels)
#  res2 <- removeSpots(arrayDataObject=arrayDataObject,exprSetArrayObject=exprSetArrayObject, spotsToBeRemoved=c("x","z"))
#  stopifnot( dim(exprs(res2[["exprSetArrayObject"]]))[1] == 3*20 )
#  res3 <- removeSpots(arrayDataObject=arrayDataObject,exprSetArrayObject=exprSetArrayObject, spotsToBeRemoved=c("ggrrrr"))
#  stopifnot( dim(exprs(res3[["exprSetArrayObject"]]))[1] == 100 )
#  stopifnot( dim(getIntensities(res3[["arrayDataObject"]]))[1] == 100 )
#	}
#
# \keyword{}
#
# \author{Andreas Buness <a.buness@dkfz.de>}
# */

removeSpots <- function(arrayDataObject, exprSetArrayObject=NULL, spotsToBeRemoved=NULL, spotIdentifier="Name"){

  if( missing(arrayDataObject) ){
     stop(" arrayDataObject is missing in removeSpots ")
   }
  if( ! class(arrayDataObject) == "arrayData" ){
    stop(" wrong class for object arrayDataObject in removeSpots ")
  }

  
  if( ! is.null(exprSetArrayObject) ){
    if( ! class(exprSetArrayObject) == "exprSetArray" ){
      stop(" wrong class for object exprSetArrayObject  in removeSpots ")
    }
    if( dim(exprs(exprSetArrayObject))[1] != dim(getIntensities(arrayDataObject))[1] ){
      stop(paste(" number of rows of exprSetArrayObject:", (dim(exprs(exprSetArrayObject))[1]), " and arrayDataObject:", (dim(getIntensities(arrayDataObject))[1])," differ in removeSpots"))
    }
    if(  dim(exprs(exprSetArrayObject))[2]/2 != dim(getIntensities(arrayDataObject))[3] ){
      stop(paste(" number of hybridisations of exprSetArrayObject:", (dim(exprs(exprSetArrayObject))[2]/2)," and of arrayDataObject:",  (dim(getIntensities(arrayDataObject))[3])," differ in removeSpots"))
    }
  }
  
  spotAttr <- getSpotAttr(arrayDataObject)

  defaultReturnValue <- list(arrayDataObject=arrayDataObject, exprSetArrayObject=exprSetArrayObject, nrOfRemovedItems=0)
  
  if( is.null(spotAttr) ){
    warning(" getSpotAttr(arrayDataObject) is NULL in removeSpots, skipped ")
    return(defaultReturnValue)
  }
  if( length(spotAttr) < 1 ){
    warning(" getSpotAttr(arrayDataObject) has length < 1 in removeSpots, skipped ")
    return(defaultReturnValue)
  }
  if(is.null(spotIdentifier)){
    warning(" spot identifier is NULL in removeSpots, skipped ")
    return(defaultReturnValue)
  }
  if( all(is.na(spotIdentifier)) |
     length(spotIdentifier) != 1 ){
    warning(" no spot identifier given in removeSpots, skipped ")
    return(defaultReturnValue)
  }

  if(is.null(spotsToBeRemoved)){
    cat(" spotsToBeRemoved is NULL in removeSpots, skipped \n")
    return(defaultReturnValue)
  }
  if( any(is.na(spotsToBeRemoved)) ){
    warning(" spotsToBeRemoved contain NA in removeSpots, skipped \n")
    return(defaultReturnValue)
  }
  if( length(spotsToBeRemoved) < 1 ){
    cat(" no spotsToBeRemoved given in removeSpots, skipped \n")
    return(defaultReturnValue)
  }
  
  if( ! (spotIdentifier %in% colnames(spotAttr)) ){
    warning(" no appropriate column in getSpotAttr(arrayDataObject) in function removeSpots, skipped ")
    return(defaultReturnValue)
  }

  
  removeItems <- sapply(spotsToBeRemoved, function(x) return(which( spotAttr[,spotIdentifier] == x)), simplify=TRUE )
  removeItems <- na.exclude(as.numeric(unlist(removeItems)))

  nrOfRemovedItems <- length(removeItems)
  if( nrOfRemovedItems == 0 ){
    newAData <- arrayDataObject
    if( ! is.null(exprSetArrayObject) ){
      newEData <- exprSetArrayObject
    }else{
      newEData <- NULL
    }
  }else{
    if(!is.null(getWeights(arrayDataObject)) ){
      newWeights <- getWeights(arrayDataObject)[-removeItems,,drop=FALSE]
    }else{
      newWeights <- NULL
    }
    newAData <- new("arrayData",
                    intensities=getIntensities(arrayDataObject)[-removeItems,,,drop=FALSE],
                    spotattr=subset(spotAttr,
                      subset=! 1:dim(spotAttr)[1] %in% removeItems),
                    weights=newWeights,
                    hybattr=getHybAttr(arrayDataObject))
    if( ! is.null(exprSetArrayObject) ){
      newEData <- exprSetArrayObject[-removeItems,]
    }else{
      newEData <- NULL
    }
  }
  
  return(list(arrayDataObject=newAData, exprSetArrayObject=newEData, nrOfRemovedItems=nrOfRemovedItems))
  
}
